open Tags

type def = { id: int; final: bool; mutable trans: (SymbolSet.t * t * t) list;
	     mutable status: status }

and t = d ref
and d = Def of def | Link of t | Undef
and status = Empty | Sample of Syntax.t | Unknown

let empty_def = { id = 0; final = false; trans = []; status = Empty }
let empty = ref (Def empty_def)

let rec repr = function
  | { contents = Link t } -> repr t
  | t -> t


exception NonDef

let rec def = function
  | { contents = Link t } -> def t
  | { contents = Def d } -> d
  | _ -> raise NonDef

let rec def_or_empty q = try def q with NonDef -> empty_def

let is_undef q = !(repr q) == Undef

let id q = (def q).id
let trans q = (def q).trans
let final q = (def q).final

let equal q1 q2 = repr q1 == repr q2
let hash q = id q

let link q1 q2 =
  let q1 = repr q1 in
  q1 := Link q2
  
let compare q1 q2 = (id q1) - (id q2)

type s = t
module StateSet = Set.Make(struct type t = s
				  let compare = compare end)

let print_id ppf q =
  try let i = id q in Format.fprintf ppf "%i" i
  with NonDef -> Format.fprintf ppf "*"

let print_def ppf def =
  if def.final then Format.fprintf ppf " %%";
  List.iter
    (fun (sset, qsub, qnext) ->
       Format.fprintf ppf " %a[%a],%a" SymbolSet.print sset
	 print_id qsub print_id qnext
    )
    def.trans

let print ppf q = Format.fprintf ppf "[%i] %a" (id q) print_def (def q)

let closures = Hashtbl.create 1024

let iter_next f q =
  List.iter (fun (_,_,next) -> f next) (trans q)

let compute_closure s =
  let accu = ref StateSet.empty in
  let rec aux s =
    if StateSet.mem s !accu then ()
    else (
      accu := StateSet.add s !accu;
      iter_next aux s
    )
  in
  aux s;
  !accu

let rec closure q =
  let i = id q in
  try Hashtbl.find closures i
  with Not_found ->
    let x = compute_closure q in
    Hashtbl.add closures i x;
    x

let between q1 q2 =
  List.filter
    (fun q -> StateSet.mem q2 (closure q))
    (StateSet.elements (closure q1))

let finals q1 =
  List.filter
    final
    (StateSet.elements (closure q1))
    

let state_id = ref 0

let mk () = ref Undef

let set q f tr =
  assert (!q == Undef);
  let tr = 
    List.filter
      (fun (sset,qsub,qnext) -> not (SymbolSet.is_empty sset))
      tr in
  if (not f) && (tr = []) then link q empty
  else
    q := Def { id = (incr state_id; !state_id); final = f; trans = tr;
	       status = (if f then Sample Syntax.Eps else Unknown) };
(*  Format.fprintf Format.std_formatter "%a@." print q; *)
  ()

let cons f tr = 
  let q = mk () in
  set q f tr;
  q

let any = 
  let q = mk () in
  set q true [ SymbolSet.any, q, q ];
  q

let empty_seq =  cons true []

let alt q1 q2 = 
  let d1 = def_or_empty q1 and d2 = def_or_empty q2 in 
  cons (d1.final || d2.final) (d1.trans @ d2.trans)

let stack = ref []

let rec destack q =
  match !stack with
    | hd::tl when hd == q -> stack := tl
    | hd::tl -> stack := tl; destack q
    | [] -> assert false

let sample_elem sset sub next = 
  let e = Syntax.Elem (SymbolSet.finite [SymbolSet.sample sset], sub) in
  match next with
    | Syntax.Eps -> e
    | next -> Syntax.Seq (e,next)

let rec sample q =
  let d = def q in
  match d.status with
    | Empty -> None
    | Sample t -> Some t
    | Unknown ->
	d.status <- Empty;
	stack := q :: !stack;
	let rec aux = function
	  | [] -> None
	  | (sset,qsub,qnext) :: rest ->
	      match sample qsub with None -> aux rest | Some sub ->
		match sample qnext with None -> aux rest | Some next ->
		  let s = sample_elem sset sub next in
		  d.status <- Sample s;
		  destack q;
		  Some s
	in
	aux d.trans
	
let sample q =
  let r = sample q in
  List.iter (fun q -> link q empty) !stack;
  stack := [];
  r

let is_empty q =
  sample q == None

let sample q =
  match sample q with
    | Some x -> x
    | None -> raise Not_found

let auto q =
  let d = def q in
  List.fold_left 
    (fun accu (sset,qsub,qnext) -> 
       if is_empty qsub || is_empty qnext then accu
       else ((sset,qsub),qnext)::accu) [] d.trans, d.final

let print_def ppf q = 
  let d = def q in
  let d = 
    { d with
	trans = 
	List.filter
	  (fun (sset,qsub,qnext) -> 
	     not (is_empty qsub) && not (is_empty qnext))
	  d.trans } in
  print_def ppf d

let rec match_forest q = function
  | [] -> (def q).final
  | (tag,sub)::next ->
      List.exists
	(fun (sset,qsub,qnext) ->
	   SymbolSet.has sset tag
	   && match_val qsub sub
	   && match_forest qnext next
	) (def q).trans

and match_val q t = match_forest q (Tree.get t)
  
