open Tags
open Expr
open Syntax

let fold_left2 f accu l1 l2 =
  List.fold_left
    (fun accu x -> List.fold_left (fun accu y -> f accu x y) accu l2)
    accu l1

module Var = struct
  type t = Syntax.t * State.t * State.t
  let compare (e1,start1,end1) (e2,start2,end2) =
    let c = Syntax.compare e1 e2 in if c !=0 then c 
    else let c = State.compare start1 start2 in if c != 0 then c
    else State.compare end1 end2

  let is_same (_,qstart,qend) = State.equal qstart qend
end


module VarList = Slist.Make(Var)
module StateList = Slist.Make(State)

module V = struct
  type t = VarList.t * StateList.t * StateList.t

  let compare (vs1,pos1,neg1) (vs2,pos2,neg2) =
    let c = VarList.compare vs1 vs2 in if c != 0 then c
    else let c = StateList.compare pos1 pos2 in if c != 0 then c
    else StateList.compare neg1 neg2

  let inter (vs1,pos1,neg1) (vs2,pos2,neg2) =
    (VarList.union vs1 vs2,
     StateList.union pos1 pos2,
     StateList.union neg1 neg2)

  let any = [], [], []

  let var v = [v], [], []
  let state q = [], [q], []

  let minus_q (vs,pos,neg) q = (vs, pos, StateList.add q neg)
  let has_neg_q (_,_,neg) q = StateList.mem q neg
  let has_q (_,pos,_) q = StateList.mem q pos

  let triv_sub (vs1,pos1,neg1) (vs2,pos2,neg2) =
    VarList.subset vs2 vs1
    && StateList.subset pos2 pos1
    && StateList.subset neg2 neg1

  let print ppf (vs,pos,neg) =
    let first = ref true in
    let sep () = if !first then first := false else Format.fprintf ppf "," in
    List.iter 
      (fun (e,qstart,qend) -> sep ();
	 Format.fprintf ppf "%a(%i->%i)" Syntax.print e
	   (State.id qstart) (State.id qend))
      vs;
    List.iter (fun q -> sep (); Format.fprintf ppf "%i" (State.id q)) pos;
    List.iter (fun q -> sep (); Format.fprintf ppf "-%i" (State.id q)) neg
end

module VMap = Map.Make(V)

let memo = ref VMap.empty

module D = struct    
  type t = { final : bool; trans : (SymbolSet.t * V.t * V.t) list }

  let any = { final = true; trans = [ SymbolSet.any, V.any, V.any ] }
  let empty = { final = false; trans = [] }

  let is_empty d = not d.final && d.trans == []

  let final accu q1 q2 = 
    if State.equal q1 q2 then { accu with final = true } else accu

  let intersect accu d1 d2 =
    { final = accu.final || (d1.final && d2.final);
      trans = 
	fold_left2
	  (fun accu (s1,sub1,next1) (s2,sub2,next2) ->
	     let s = SymbolSet.inter s1 s2 in
	     if SymbolSet.is_empty s 
	     then accu 
	     else (s, V.inter sub1 sub2, V.inter next1 next2) :: accu
	  )
	  accu.trans
	  d1.trans d2.trans }

  let union d1 d2 = { final = d1.final || d2.final;
		      trans = d1.trans @ d2.trans }

  let inter = intersect empty

  let tag sset = 
    { final = false; trans = [ sset, V.any, V.any ] }

  let cstr q = 
    { final = State.final q;
      trans = 
	List.map 
	  (fun (sset,qsub,qnext) -> sset, V.state qsub, V.state qnext)
	  (State.trans q) }

  let ncstr q =
    let rec aux accu sset sub next = function
      | [] -> (sset,sub,next) :: accu
      | (sset',qsub,qnext) :: rest ->
	  let i = SymbolSet.inter sset sset' in
	  if SymbolSet.is_empty i || V.has_neg_q sub qsub 
	    || V.has_neg_q next qnext
	  then aux accu sset sub next rest
	  else
	    let accu = 
	      if V.has_q sub qsub then accu
	      else aux accu i (V.minus_q sub qsub) next rest in
	    let accu = 
	      if V.has_q next qnext then accu
	      else aux accu i sub (V.minus_q next qnext) rest in
	    let d = SymbolSet.diff sset sset' in
	    if SymbolSet.is_empty d then accu
	    else aux accu d sub next rest
    in
    { final = not (State.final q);
      trans = aux [] SymbolSet.any V.any V.any (State.trans q) }

  let sub accu v =
    { accu with
	trans = (SymbolSet.any, V.var v, V.any) :: accu.trans }
  let next accu v =
    { accu with
	trans = (SymbolSet.any, V.any, V.var v) :: accu.trans } 

  let triv_sub (sset1,sub1,next1) (sset2,sub2,next2) =
    SymbolSet.subset sset1 sset2 
    && V.triv_sub sub1 sub2
    && V.triv_sub next1 next2
      
  let rec filter_triv_sub accu = function
    | [] -> accu
    | h::t ->
	if List.exists (triv_sub h) accu || List.exists (triv_sub h) t 
	then filter_triv_sub accu t
	else filter_triv_sub (h::accu) t

  let clean t = { t with trans = filter_triv_sub [] t.trans }

end

(*
let rec descr e qstart qend accu sset pos neg =
  let accu' = descr_aux e qstart qend accu sset pos neg in
  Printf.eprintf "descr %i->%i => %i,%i\n" (State.id qstart) (State.id qend) (List.length accu) (List.length accu');
  accu'
*)

  
let rec descr e qstart qend accu =
  if State.equal qstart State.any && State.equal qend State.any then D.any
  else match e with
    | Eps -> if State.equal qstart qend then D.any else accu
    | Seq (e1,e2) ->
	List.fold_left
	  (fun accu q ->
	     let s1 = descr e1 qstart q D.empty in
	     if D.is_empty s1 then accu
	     else D.intersect accu s1 (descr' e2 q qend)
	  )
	  accu
	  (State.between qstart qend)
(* TODO: special case when length e1 = 1 *)
    | Elem (symb, e) ->
	if SymbolSet.is_singleton symb then 
	  let symb = SymbolSet.sample symb in
	  List.fold_left
	    (fun accu (sset, qsub, qnext) ->
	       if State.equal qnext qend && SymbolSet.has sset symb then
		 List.fold_left
		   (fun accu qfin -> descr e qsub qfin accu)
		   accu
		   (State.finals qsub)
	       else accu
	    )
	    accu
	    (State.trans qstart)
	else 
	  List.fold_left
	    (fun accu (sset, qsub, qnext) ->
	       if State.equal qnext qend then
		 let sset = D.tag sset in
		 List.fold_left
		   (fun accu qfin -> 
		      D.intersect accu (descr' e qsub qfin) sset)
		   accu
		   (State.finals qsub)
	       else accu
	    )
	    (D.final accu qstart qend)
	    (State.trans qstart)
    | Test (cstr, ethen, eelse) ->
	let q = Regexp.compile cstr in
	let accu = D.intersect accu (descr' ethen qstart qend) (D.cstr q) in
	let accu = D.intersect accu (descr' eelse qstart qend) (D.ncstr q) in
	accu
    | Star e as star -> descr (Seq(e,Next star)) qstart qend accu
    | Sub e -> D.sub (D.final accu qstart qend) (e,qstart,qend)
    | Next e -> D.next (D.final accu qstart qend) (e,qstart,qend)
    | Name x ->
	descr (Syntax.def x) qstart qend accu
    | _ -> assert false


and descr' e qstart qend = descr e qstart qend D.empty

let descr_var (e,qstart,qend) = descr' e qstart qend

let triv_sub (sset1,sub1,next1) (sset2,sub2,next2) =
  SymbolSet.subset sset1 sset2 
  && V.triv_sub sub1 sub2
  && V.triv_sub next1 next2

let rec filter_triv_sub accu = function
  | [] -> accu
  | h::t ->
      if List.exists (triv_sub h) accu || List.exists (triv_sub h) t 
      then filter_triv_sub accu t
      else filter_triv_sub (h::accu) t

let rec descr_varset q (vs,pos,neg) =
  let t = List.fold_left (fun accu v -> D.inter accu (descr_var v)) D.any vs in
  let t = List.fold_left (fun accu q -> D.inter accu (D.cstr q)) t pos in
  let t = List.fold_left (fun accu q -> D.inter accu (D.ncstr q)) t neg in
  mk q t
 
and mk q t =
  let t = D.clean t in
  State.set q t.D.final 
    (List.map (fun (s, l1, l2) -> (s, state_of l1, state_of l2)) t.D.trans)


and state_of (vs,pos,neg) =
  let vs = 
    List.filter 
      (fun (_,qstart,qend) -> qstart != State.any || qend != State.any) vs in
  match vs,pos,neg with
  | [], [q], [] -> q
  | [], [], [] -> State.any
  | l ->
      try VMap.find l !memo
      with Not_found ->
	let q = State.mk () in
	memo := VMap.add l q !memo;
	descr_varset q l;
(*	Format.fprintf Format.std_formatter "State for %a: %i@." V.print l 
	  (State.id q); *)
	q

let infer e q =
  let r = State.mk () in
  mk r
    (List.fold_left D.union D.empty 
       (List.map (fun qend -> descr_var (e,q,qend)) (State.finals q)));
  r

  
let diff q1 q2 =
  state_of ([], [q1], [q2])
